"""Contain tab base classes."""

from __future__ import annotations

import logging
from abc import ABCMeta
from collections import deque
from functools import partial
from typing import TYPE_CHECKING

from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.layout.containers import Window, WindowAlign
from prompt_toolkit.layout.controls import FormattedTextControl

from euporie.core.comm.registry import open_comm
from euporie.core.commands import add_cmd
from euporie.core.completion import KernelCompleter
from euporie.core.config import add_setting
from euporie.core.current import get_app
from euporie.core.filters import kernel_tab_has_focus, tab_has_focus
from euporie.core.history import KernelHistory
from euporie.core.kernel import Kernel, MsgCallbacks
from euporie.core.key_binding.registry import (
    register_bindings,
)
from euporie.core.suggest import HistoryAutoSuggest
from euporie.core.utils import run_in_thread_with_context

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Callable, Sequence

    from prompt_toolkit.auto_suggest import AutoSuggest
    from prompt_toolkit.completion.base import Completer
    from prompt_toolkit.history import History
    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.app import BaseApp
    from euporie.core.comm.base import Comm
    from euporie.core.widgets.status_bar import StatusBarFields

log = logging.getLogger(__name__)


class Tab(metaclass=ABCMeta):
    """Base class for interface tabs."""

    _registry: set[type[Tab]] = set()
    name: str | None = None
    weight: int = 0
    mime_types: set[str] = set()
    file_extensions: set[str] = set()

    container: AnyContainer

    def __init_subclass__(cls, *args: Any, **kwargs: Any) -> None:
        """Compile a registry of named tabs."""
        super().__init_subclass__(**kwargs)
        if cls.name:
            Tab._registry.add(cls)

    def __init__(self, app: BaseApp, path: Path | None = None) -> None:
        """Call when the tab is created."""
        self.app = app
        self.path = path
        self.container = Window(
            FormattedTextControl([("fg:#888888", "\nLoading…")], focusable=True),
            align=WindowAlign.CENTER,
        )

        self.dirty = False
        self.saving = False

    @property
    def title(self) -> str:
        """Return the tab title."""
        return ""

    def reset(self) -> "None":  # noqa B027
        """Reet the state of the tab."""
        pass

    def close(self, cb: Callable | None = None) -> None:
        """Close a tab with a callback.

        Args:
            cb: A function to call after the tab is closed.

        """
        # Run callback
        if callable(cb):
            cb()

    def focus(self) -> None:
        """Focus the tab (or make it visible)."""
        self.app.focus_tab(self)

    def _save(self, path: Path | None = None, cb: Callable | None = None) -> None:
        """Perform the file save in a background thread."""
        run_in_thread_with_context(self.save, path, cb)

    def save(self, path: Path | None = None, cb: Callable | None = None) -> None:
        """Save the current notebook."""
        raise NotImplementedError

    def __pt_status__(self) -> StatusBarFields | None:
        """Return a list of statusbar field values shown then this tab is active."""
        return ([], [])

    def __pt_container__(self) -> AnyContainer:
        """Return the main container object."""
        return self.container

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(filter=tab_has_focus, title="Refresh the current tab")
    def _refresh_tab() -> None:
        """Reload the tab contents and reset the tab."""
        if (tab := get_app().tab) is not None:
            tab.reset()

    # Depreciated v2.5.0
    @staticmethod
    @add_cmd(filter=tab_has_focus, title="Reset the current tab")
    def _reset_tab() -> None:
        log.warning(
            "The `reset-tab` command was been renamed to `refresh-tab` in v2.5.0,"
            " and will be removed in a future version"
        )
        Tab._refresh_tab()

    @staticmethod
    @add_cmd(filter=tab_has_focus)
    def _save_file() -> None:
        """Save the current file."""
        if (tab := get_app().tab) is not None:
            try:
                tab._save()
            except NotImplementedError:
                pass

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.tabs.base.Tab": {
                "save-file": "c-s",
                "refresh-tab": "f5",
            }
        }
    )


class KernelTab(Tab, metaclass=ABCMeta):
    """A Tab which connects to a kernel."""

    kernel: Kernel
    kernel_language: str
    _metadata: dict[str, Any]
    bg_init = True

    default_callbacks: MsgCallbacks
    allow_stdin: bool

    def __init__(
        self,
        app: BaseApp,
        path: Path | None = None,
        kernel: Kernel | None = None,
        comms: dict[str, Comm] | None = None,
        use_kernel_history: bool = False,
        connection_file: Path | None = None,
    ) -> None:
        """Create a new instance of a tab with a kernel."""
        # Init tab
        super().__init__(app, path)

        if self.bg_init:
            # Load kernel in a background thread
            run_in_thread_with_context(
                partial(
                    self.init_kernel, kernel, comms, use_kernel_history, connection_file
                )
            )
        else:
            self.init_kernel(kernel, comms, use_kernel_history, connection_file)

    def pre_init_kernel(self) -> None:
        """Run stuff before the kernel is loaded."""
        pass

    def post_init_kernel(self) -> None:
        """Run stuff after the kernel is loaded."""
        pass

    def init_kernel(
        self,
        kernel: Kernel | None = None,
        comms: dict[str, Comm] | None = None,
        use_kernel_history: bool = False,
        connection_file: Path | None = None,
    ) -> None:
        """Set up the tab's kernel and related components."""
        self.pre_init_kernel()

        self.kernel_queue: deque[Callable] = deque()

        if kernel:
            self.kernel = kernel
            self.kernel.default_callbacks = self.default_callbacks
        else:
            self.kernel = Kernel(
                kernel_tab=self,
                allow_stdin=self.allow_stdin,
                default_callbacks=self.default_callbacks,
                connection_file=connection_file,
            )
        self.comms: dict[str, Comm] = comms or {}  # The client-side comm states
        self.completer: Completer = KernelCompleter(self.kernel)
        self.use_kernel_history = use_kernel_history
        self.history: History = (
            KernelHistory(self.kernel) if use_kernel_history else InMemoryHistory()
        )
        self.suggester: AutoSuggest = HistoryAutoSuggest(self.history)

        self.post_init_kernel()

    def close(self, cb: Callable | None = None) -> None:
        """Shut down kernel when tab is closed."""
        if hasattr(self, "kernel"):
            self.kernel.shutdown()
        super().close(cb)

    def interrupt_kernel(self) -> None:
        """Interrupt the current `Notebook`'s kernel."""
        self.kernel.interrupt()

    def restart_kernel(self, cb: Callable | None = None) -> None:
        """Restart the current `Notebook`'s kernel."""

        def _cb(result: dict[str, Any]) -> None:
            if callable(cb):
                cb()

        if confirm := self.app.dialogs.get("confirm"):
            confirm.show(
                message="Are you sure you want to restart the kernel?",
                cb=partial(self.kernel.restart, cb=_cb),
            )
        else:
            self.kernel.restart(cb=_cb)

    def kernel_started(self, result: dict[str, Any] | None = None) -> None:
        """Task to run when the kernel has started."""
        # Check kernel has not failed
        if not self.kernel_name or self.kernel.missing:
            if not self.kernel_name:
                msg = "No kernel selected"
            else:
                msg = f"Kernel '{self.kernel_display_name}' not installed"
            self.change_kernel(
                msg=msg,
                startup=True,
            )

        elif self.kernel.status == "error":
            self.report_kernel_error(self.kernel.error)

        else:
            # Wait for an idle kernel
            if self.kernel.status != "idle":
                self.kernel.wait_for_status("idle")

            # Load widget comm info
            # self.kernel.comm_info(target_name="jupyter.widget")

            # Load kernel info
            self.kernel.info(set_kernel_info=self.set_kernel_info)

            # Load kernel history
            if self.use_kernel_history:
                self.app.create_background_task(self.load_history())

            # Run queued kernel tasks when the kernel is idle
            log.debug("Running %d kernel tasks", len(self.kernel_queue))
            while self.kernel_queue:
                self.kernel_queue.popleft()()

        self.app.invalidate()

    def report_kernel_error(self, error: Exception | None) -> None:
        """Report a kernel error to the user."""
        log.debug("Kernel error", exc_info=error)

    async def load_history(self) -> None:
        """Load kernel history."""
        try:
            await self.history.load().__anext__()
        except StopAsyncIteration:
            pass

    @property
    def metadata(self) -> dict[str, Any]:
        """Return a dictionary to hold notebook / kernel metadata."""
        return self._metadata

    @property
    def kernel_name(self) -> str:
        """Return the name of the kernel defined in the notebook JSON."""
        return self.metadata.get("kernelspec", {}).get(
            "name", self.app.config.kernel_name
        )

    @kernel_name.setter
    def kernel_name(self, value: str) -> None:
        """Return the name of the kernel defined in the notebook JSON."""
        self.metadata.setdefault("kernelspec", {})["name"] = value

    @property
    def language(self) -> str:
        """Return the name of the kernel defined in the notebook JSON."""
        return self.metadata.get("kernelspec", {}).get("language")

    @property
    def kernel_display_name(self) -> str:
        """Return the display name of the kernel defined in the notebook JSON."""
        return self.metadata.get("kernelspec", {}).get("display_name", self.kernel_name)

    @property
    def kernel_lang_file_ext(self) -> str:
        """Return the display name of the kernel defined in the notebook JSON."""
        return self.metadata.get("language_info", {}).get("file_extension", ".py")

    def set_kernel_info(self, info: dict) -> None:
        """Handle kernel info requests."""
        self.metadata["language_info"] = info.get("language_info", {})

    def change_kernel(self, msg: str | None = None, startup: bool = False) -> None:
        """Prompt the user to select a new kernel."""
        kernel_specs = self.kernel.specs

        # Warn the user if no kernels are installed
        if not kernel_specs:
            if startup and "no-kernels" in self.app.dialogs:
                self.app.dialogs["no-kernels"].show()
            return

        # Automatically select the only kernel if there is only one
        if startup and len(kernel_specs) == 1:
            self.kernel.change(list(kernel_specs)[0])
            return

        self.app.dialogs["change-kernel"].show(
            tab=self, message=msg, kernel_specs=kernel_specs
        )

    def comm_open(self, content: dict, buffers: Sequence[bytes]) -> None:
        """Register a new kernel Comm object in the notebook."""
        comm_id = str(content.get("comm_id"))
        self.comms[comm_id] = open_comm(
            comm_container=self, content=content, buffers=buffers
        )

    def comm_msg(self, content: dict, buffers: Sequence[bytes]) -> None:
        """Respond to a Comm message from the kernel."""
        comm_id = str(content.get("comm_id"))
        if comm := self.comms.get(comm_id):
            comm.process_data(content.get("data", {}), buffers)

    def comm_close(self, content: dict, buffers: Sequence[bytes]) -> None:
        """Close a notebook Comm."""
        comm_id = content.get("comm_id")
        if comm_id in self.comms:
            del self.comms[comm_id]

    # ################################### Commands ####################################

    @staticmethod
    @add_cmd(filter=kernel_tab_has_focus)
    def _change_kernel() -> None:
        """Change the notebook's kernel."""
        if isinstance(kt := get_app().tab, KernelTab):
            kt.change_kernel()

    # ################################### Settings ####################################

    add_setting(
        name="kernel_name",
        flags=["--kernel-name", "--kernel"],
        type_=str,
        help_="The name of the kernel to start by default",
        default="python3",
        description="""
            The name of the kernel selected automatically by the console app or in new
            notebooks. If set to an empty string, the user will be asked which kernel
            to launch.
        """,
    )

    add_setting(
        name="record_cell_timing",
        title="cell timing recording",
        flags=["--record-cell-timing"],
        type_=bool,
        help_="Should timing data be recorded in cell metadata.",
        default=False,
        schema={
            "type": "boolean",
        },
        description="""
            When set, execution timing data will be recorded in cell metadata.
        """,
    )



"""Allow access to the current running application."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from prompt_toolkit.application.current import get_app as ptk_get_app

if TYPE_CHECKING:
    from euporie.notebook.app import NotebookApp


def get_app() -> NotebookApp:
    """Get the current application."""
    return cast("NotebookApp", ptk_get_app())



"""Define collections of generic key-bindings which do not belong to widgets."""

from euporie.core.key_binding.bindings import basic, completion, micro, mouse

__all__ = ["basic", "completion", "micro", "mouse"]



"""Define the base class for a Comm object and it's representation."""

from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING
from weakref import WeakKeyDictionary

from euporie.core.current import get_app
from euporie.core.widgets.display import Display

if TYPE_CHECKING:
    from typing import Any, Callable, Mapping, Sequence

    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.kernel import Kernel
    from euporie.core.tabs.base import KernelTab
    from euporie.core.widgets.cell_outputs import OutputParent


log = logging.getLogger(__name__)


class CommView:
    """Hold a container and methods to update its attributes."""

    def __init__(
        self,
        container: AnyContainer,
        setters: Mapping[str, Callable[..., None]] | None = None,
    ) -> None:
        """Create a new instance of the Comm vieew.

        Args:
            container: The main container to display when this Comm is to be visualised
            setters: A dictionary mapping state names to callback functions which should
                be called when the named state value changed.
        """
        self.container = container
        self.setters: dict[str, Callable[..., None]] = dict(setters or {})
        self.kernel: Kernel | None = None

    def update(self, changes: dict[str, Any]) -> None:
        """Update the view to reflect changes in the Comm.

        Calls any setter functions defined for the changed keys with the changed values
        as arguments.

        Args:
            changes: A dictionary mapping changed key names to new values.

        """
        for key, value in changes.items():
            if setter := self.setters.get(key):
                setter(value)

    def __pt_container__(self) -> AnyContainer:
        """Return the widget's container for display."""
        return self.container


class Comm(metaclass=ABCMeta):
    """A base-class for all comm objects, which support syncing traits with the kernel."""

    def __init__(
        self,
        comm_container: KernelTab,
        comm_id: str,
        data: dict,
        buffers: Sequence[bytes],
    ) -> None:
        """Create a new instance of the Comm.

        Args:
            comm_container: The container this Comm belongs to
            comm_id: The ID of the Comm
            data: The data field from the ``comm_open`` message
            buffers: The buffers field from the ``comm_open`` message

        """
        self.comm_container = comm_container
        self.comm_id = comm_id
        self.data: dict[str, Any] = {}
        self.buffers: Sequence[bytes] = []
        self.views: WeakKeyDictionary[CommView, OutputParent] = WeakKeyDictionary()
        self.process_data(data, buffers)

    @abstractmethod
    def process_data(self, data: dict, buffers: Sequence[bytes]) -> None:
        """Process a comm_msg data / buffers."""

    def _get_embed_state(self) -> dict:
        return {}

    def create_view(self, parent: OutputParent) -> CommView:
        """Create a new :class:`CommView` for this Comm."""
        return CommView(Display("[Object cannot be rendered]", format_="ansi"))

    def new_view(self, parent: OutputParent) -> CommView:
        """Create and register a new :class:`CommView` for this Comm."""
        view = self.create_view(parent)
        self.views[view] = parent
        return view

    def update_views(self, changes: dict) -> None:
        """Update all the active views of this Comm."""
        for view, parent in self.views.items():
            view.update(changes)
            parent.refresh(now=False)
        get_app().invalidate()


class UnimplementedComm(Comm):
    """Represent a Comm object which is not implemented in euporie.core."""

    def process_data(self, data: dict, buffers: Sequence[bytes]) -> None:
        """Doe nothing when data is received."""
        self.data = data
        self.buffers = buffers



"""Contain functions which convert data to markdown format."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from euporie.core.convert.core import register
from euporie.core.convert.utils import have_modules

if TYPE_CHECKING:
    from pathlib import Path

log = logging.getLogger(__name__)

_HTML2TEXT_TABLE_RE = r"(?:(?:.*\|)+[^|]*?(?:\n|$))+"


@register(
    from_="html",
    to="markdown",
    filter_=have_modules("html2text"),
)
async def html_to_markdown_py_html2text(
    data: str | bytes,
    width: int | None = None,
    height: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
    initial_format: str = "",
) -> str:
    """Convert HTML to markdown tables using :py:mod:`html2text`."""
    import re

    from html2text import HTML2Text

    parser = HTML2Text(bodywidth=0)
    markup = data.decode() if isinstance(data, bytes) else data
    result = parser.handle(markup)

    # Fix for html2text issue with empty first cells in table rows
    remaining_result = result
    replaced = ""
    for match in re.finditer(_HTML2TEXT_TABLE_RE, result, re.MULTILINE):
        # Add string before the table
        replaced += remaining_result[
            : match.start() - (len(result) - len(remaining_result))
        ]
        remaining_result = result[match.end() :]
        table = result[match.start() : match.end()]
        table_rows = table.strip().split("\n")
        row_lens = [len(row.split("|")) for row in table_rows]
        max_row_len = max(row_lens)
        table = "\n".join(
            "|" * (max_row_len - row_lens[i] + 1) + row
            for i, row in enumerate(table_rows)
        )
        replaced += table + "\n"
    replaced += remaining_result
    result = replaced

    return result


@register(
    from_="html",
    to="markdown",
    filter_=have_modules("mtable", "html5lib"),
)
async def html_to_markdown_py_mtable(
    data: str | bytes,
    width: int | None = None,
    height: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
    initial_format: str = "",
) -> str:
    """Convert HTML tables to markdown tables using :py:mod:`mtable`."""
    from mtable import MarkupTable

    markup = data.decode() if isinstance(data, bytes) else data
    return "\n\n".join([table.to_md() for table in MarkupTable.from_html(markup)])



"""Define basic key-bindings for entering text."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.filters import (
    buffer_has_focus,
)
from prompt_toolkit.key_binding import ConditionalKeyBindings

from euporie.core.commands import add_cmd
from euporie.core.filters import (
    micro_mode,
    replace_mode,
)
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.key_binding.utils import if_no_repeat

if TYPE_CHECKING:
    from prompt_toolkit.key_binding import KeyBindingsBase, KeyPressEvent

    from euporie.core.config import Config

log = logging.getLogger(__name__)


class TextEntry:
    """Basic key-bindings for text entry."""


# Register default bindings for micro edit mode
register_bindings(
    {
        "euporie.core.key_binding.bindings.basic.TextEntry": {
            "type-key": "<any>",
        },
    }
)


def load_basic_bindings(config: Config | None = None) -> KeyBindingsBase:
    """Load basic key-bindings for text entry."""
    return ConditionalKeyBindings(
        load_registered_bindings(
            "euporie.core.key_binding.bindings.basic.TextEntry", config=config
        ),
        micro_mode,
    )


# Commands


@add_cmd(
    filter=buffer_has_focus,
    save_before=if_no_repeat,
    hidden=True,
)
def type_key(event: KeyPressEvent) -> None:
    """Enter a key."""
    # Do not insert escape sequences
    if not event.data.startswith("\x1b["):
        event.current_buffer.insert_text(
            event.data * event.arg, overwrite=replace_mode()
        )



"""Sub-module concerned with the conversion of data formats."""

from euporie.core.convert.formats import (
    ansi,
    base64,
    ft,
    html,
    jpeg,
    markdown,
    pdf,
    pil,
    png,
    rich,
    sixel,
)

__all__ = [
    "ansi",
    "base64",
    "jpeg",
    "html",
    "markdown",
    "ft",
    "pdf",
    "pil",
    "png",
    "rich",
    "sixel",
]



"""Contain main format conversion function."""

from __future__ import annotations

import asyncio
import hashlib
import logging
import mimetypes
import threading
from functools import lru_cache
from typing import TYPE_CHECKING, NamedTuple

from prompt_toolkit.cache import FastDictCache, SimpleCache
from prompt_toolkit.filters import to_filter
from upath import UPath
from upath.implementations.http import HTTPPath

# from euporie.core.cache import cache
from euporie.core.path import DataPath

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any, Callable, Iterable

    from prompt_toolkit.filters import Filter, FilterOrBool

log = logging.getLogger(__name__)


MIME_FORMATS = {
    "image/svg+xml": "svg",
    "image/png": "png",
    "image/jpeg": "jpeg",
    "image/gif": "gif",
    "application/pdf": "pdf",
    "text/html": "html",
    "text/latex": "latex",
    "application/x-latex": "latex",
    "text/markdown": "markdown",
    "text/x-markdown": "markdown",
    "text/*": "ansi",
    "stream/std*": "ansi",
    "*": "ansi",
}


BASE64_FORMATS = {"png", "jpeg", "pdf", "gif"}

ERROR_OUTPUTS = {
    "ansi": b"(Format Conversion Error)",
    "ft": [("fg:white bg:darkred", "(Format Conversion Error)")],
}


_IO_THREAD: list[threading.Thread | None] = [None]  # dedicated conversion IO thread
_LOOP: list[asyncio.AbstractEventLoop | None] = [
    None
]  # global event loop for conversion


@lru_cache
def get_mime(path: Path | str) -> str | None:
    """Attempt to determine the mime-type of a path."""
    if isinstance(path, str):
        path = UPath(path)
    try:
        path = path.resolve()
    except Exception:
        log.debug("Cannot resolve '%s'", path)

    mime = None

    # Read from path of data URI
    if isinstance(path, DataPath):
        mime = path._mime

    # If we have a web-address, nsure we have a url
    # Check http-headers and nsure we have a url
    if not mime and isinstance(path, HTTPPath) and path._url is not None:
        from fsspec.asyn import sync

        # Get parsed url
        url = path._url.geturl()
        # Get the fsspec fs
        fs = path._accessor._fs
        # Ensure we have a session
        session = sync(fs.loop, fs.set_session)
        # Use HEAD requests if the server allows it, falling back to GETs
        for method in (session.head, session.get):
            r = sync(fs.loop, method, url, allow_redirects=True)
            try:
                r.raise_for_status()
            except Exception:
                log.debug("Request failed: %s", r)
                continue
            else:
                content_type = r.headers.get("Content-Type")
                if content_type is not None:
                    mime = content_type.partition(";")[0]
                    break

    # Try using magic
    if not mime:
        try:
            import magic
        except ModuleNotFoundError:
            pass
        else:
            try:
                with path.open(mode="rb") as f:
                    mime = magic.from_buffer(f.read(2048), mime=True)
            except FileNotFoundError:
                pass

    # Guess from file-extension
    if not mime and path.suffix:
        # Check for Jupyter notebooks by extension
        if path.suffix == ".ipynb":
            return "application/x-ipynb+json"
        else:
            mime, _ = mimetypes.guess_type(path)

    return mime


@lru_cache
def get_format(path: Path | str, default: str = "") -> str:
    """Attempt to guess the format of a path."""
    if isinstance(path, str):
        path = UPath(path)
    if not default:
        if isinstance(path, HTTPPath):
            default = "html"
        else:
            default = "ansi"
    mime = get_mime(path)
    return MIME_FORMATS.get(mime, default) if mime else default


class Converter(NamedTuple):
    """Hold a conversion function and its weight."""

    func: Callable
    filter_: Filter
    weight: int = 1


converters: dict[str, dict[str, list[Converter]]] = {}

_CONVERSION_CACHE: SimpleCache = SimpleCache(maxsize=2048)
_FILTER_CACHE: SimpleCache = SimpleCache()


def register(
    from_: Iterable[str] | str,
    to: str,
    filter_: FilterOrBool = True,
    weight: int = 1,
) -> Callable:
    """Add a converter to the centralized format conversion system."""
    if isinstance(from_, str):
        from_ = (from_,)

    def decorator(func: Callable) -> Callable:
        if to not in converters:
            converters[to] = {}
        for from_format in from_:
            if from_format not in converters[to]:
                converters[to][from_format] = []
            converters[to][from_format].append(
                Converter(func=func, filter_=to_filter(filter_), weight=weight)
            )
        return func

    return decorator


def find_route(from_: str, to: str) -> list | None:
    """Find the shortest conversion path between two formats."""
    if from_ == to:
        return [from_]

    chains = []

    def find(start: str, chain: list[str]) -> None:
        if chain[0] == start:
            chains.append(chain)
        sources: dict[str, list[Converter]] = converters.get(chain[0], {})
        for link in sources:
            if link not in chain:
                if any(
                    _FILTER_CACHE.get((conv,), conv.filter_)
                    for conv in sources.get(link, [])
                ):
                    find(start, [link, *chain])

    find(from_, [to])

    if chains:
        # Find chain with shortest weighted length
        return sorted(
            chains,
            key=lambda chain: sum(
                [
                    min(
                        [
                            conv.weight
                            for conv in converters.get(step_b, {}).get(step_a, [])
                            if _FILTER_CACHE.get((conv,), conv.filter_)
                        ]
                    )
                    for step_a, step_b in zip(chain, chain[1:])
                ]
            ),
        )
    else:
        return None


_CONVERTOR_ROUTE_CACHE: FastDictCache[tuple[str, str], list | None] = FastDictCache(
    find_route
)


async def _do_conversion(
    data: str | bytes,
    from_: str,
    to: str,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
) -> Any:
    if from_ == to:
        return data
    routes = _CONVERTOR_ROUTE_CACHE[(from_, to)]
    log.debug("Converting from '%s' to '%s' using route: %s", from_, to, routes)
    if not routes:
        raise NotImplementedError(f"Cannot convert from `{from_}` to `{to}`")
    output: Any = data
    for route in routes:
        for stage_a, stage_b in zip(route, route[1:]):
            # Find converter with lowest weight
            func = sorted(
                [
                    conv
                    for conv in converters[stage_b][stage_a]
                    if _FILTER_CACHE.get((conv,), conv.filter_)
                ],
                key=lambda x: x.weight,
            )[0].func
            # POSSIBLE TODO - Add intermediate steps to the cache
            try:
                output = await func(output, cols, rows, fg, bg, path, from_)
            except Exception:
                log.exception("An error occurred during format conversion")
                output = None
            if output is None:
                log.error(
                    "Failed to convert `%s` from `%s`"
                    " to `%s` using route `%s` at stage `%s`",
                    data.__repr__()[:10],
                    from_,
                    to,
                    route,
                    stage_b,
                )
                # Try the next route on error
                break
        else:
            # If this route succeeded, stop trying routes
            break
    return output


async def _convert(
    data: str | bytes,
    from_: str,
    to: str,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
) -> Any:
    """Perform conversion asynchronously, caching the result."""
    try:
        data_hash = hashlib.sha1(  # noqa S324
            data if isinstance(data, bytes) else data.encode()
        ).hexdigest()
    except TypeError:
        log.warning("Cannot hash %s", data)
        key = None
    else:
        key = (data_hash, from_, to, cols, rows, fg, bg, path)

    # We need to do a bit of juggling with the cache as SimpleCache does not support
    # using async coroutines as the getter function
    if key:
        if key in _CONVERSION_CACHE._keys:
            data = _CONVERSION_CACHE._data[key]
        else:
            _data = await _do_conversion(data, from_, to, cols, rows, fg, bg, path)
            # Pass through cache check function to keep cache size limited
            data = _CONVERSION_CACHE.get(key, lambda: _data)
    else:
        # If we can't hash the data, don't worry about caching the conversion
        data = await _do_conversion(data, from_, to, cols, rows, fg, bg, path)

    return data


def convert(
    data: Any,
    from_: str,
    to: str,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    path: Path | None = None,
) -> Any | None:
    """Convert between formats."""

    async def _runner() -> Any:
        try:
            result = await _convert(data, from_, to, cols, rows, fg, bg, path)
        except NotImplementedError:
            log.warning("Conversion not possible")
            result = None
        except Exception:
            log.exception("Error")
        return result

    future = asyncio.run_coroutine_threadsafe(_runner(), get_loop())
    data = future.result()

    if data is None:
        data = ERROR_OUTPUTS.get(to, b"(Conversion Error)")

    return data


def get_loop() -> asyncio.AbstractEventLoop:
    """Create or return the conversion IO loop.

    The loop will be running on a separate thread.
    """
    if _LOOP[0] is None:
        loop = asyncio.new_event_loop()
        _LOOP[0] = loop
        thread = threading.Thread(
            target=loop.run_forever, name="EuporieConvertIO", daemon=True
        )
        thread.start()
        _IO_THREAD[0] = thread
    assert _LOOP[0] is not None
    # Check we are not already in the conversion event loop
    try:
        running_loop = asyncio.get_running_loop()
    except RuntimeError:
        running_loop = None
    if _LOOP[0] is running_loop:
        raise NotImplementedError(
            "Cannot call `convert` from the conversion event loop"
        )
    return _LOOP[0]



"""Define the global search toolbar and related search functions."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from prompt_toolkit.filters.app import is_searching
from prompt_toolkit.key_binding.vi_state import InputMode
from prompt_toolkit.layout.controls import BufferControl, SearchBufferControl
from prompt_toolkit.search import SearchDirection
from prompt_toolkit.selection import SelectionState
from prompt_toolkit.widgets import SearchToolbar as PtkSearchToolbar

from euporie.core.commands import add_cmd
from euporie.core.current import get_app
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)

if TYPE_CHECKING:
    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.formatted_text.base import AnyFormattedText

log = logging.getLogger(__name__)


class SearchBar(PtkSearchToolbar):
    """Search mode.

    A search toolbar with custom style and text.
    """

    def __init__(
        self,
        search_buffer: Buffer | None = None,
        vi_mode: bool = False,
        text_if_not_searching: AnyFormattedText = "",
        forward_search_prompt: AnyFormattedText = "I-search: ",
        backward_search_prompt: AnyFormattedText = "I-search backward: ",
        ignore_case: FilterOrBool = False,
    ) -> None:
        """Create a new search bar instance."""
        super().__init__(
            text_if_not_searching="",
            forward_search_prompt=[
                ("class:search-toolbar.title", " Find: "),
                ("", " "),
            ],
            backward_search_prompt=[
                ("class:search-toolbar.title", " Find (up): "),
                ("", " "),
            ],
        )
        self.control.key_bindings = load_registered_bindings(
            "euporie.core.widgets.search_bar.SearchBar"
        )

    register_bindings(
        {
            "euporie.core.app.BaseApp": {
                "find": ["c-f", "f3", "f7"],
                "find-next": "c-g",
                "find-previous": "c-p",
            },
            "euporie.core.widgets.search_bar.SearchBar": {
                "accept-search": "enter",
                "stop-search": "escape",
            },
        }
    )


def start_global_search(
    buffer_control: BufferControl | None = None,
    direction: SearchDirection = SearchDirection.FORWARD,
) -> None:
    """Start a search through all searchable `buffer_controls` in the layout."""
    app = get_app()
    layout = app.layout
    current_control = layout.current_control
    # Find the search buffer control
    if app.search_bar is not None:
        search_buffer_control = app.search_bar.control
    elif (
        isinstance(current_control, BufferControl)
        and current_control.search_buffer_control is not None
    ):
        search_buffer_control = current_control.search_buffer_control
    else:
        return
    # Find all searchable controls
    searchable_controls: list[BufferControl] = []
    next_control_index = 0
    for control in layout.find_all_controls():
        # Find the index of the next searchable control so we can link the search
        # control to it if the currently focused control is not searchable. This is so
        # that the next searchable control can be focused when search is completed.
        if control == current_control:
            next_control_index = len(searchable_controls)
        # Only save the control if it is searchable
        if (
            isinstance(control, BufferControl)
            and control.search_buffer_control == search_buffer_control
        ):
            # Set its search direction
            control.search_state.direction = direction
            # Add it to our list
            searchable_controls.append(control)

    # Stop the search if we did not find any searchable controls
    if not searchable_controls:
        return

    # If the current control is searchable, link it
    if current_control in searchable_controls:
        assert isinstance(current_control, BufferControl)
        layout.search_links[search_buffer_control] = current_control
    else:
        # otherwise use the next after the currently selected control
        layout.search_links[search_buffer_control] = searchable_controls[
            next_control_index % len(searchable_controls)
        ]
    # Make sure to focus the search BufferControl
    layout.focus(search_buffer_control)
    # If we're in Vi mode, make sure to go into insert mode.
    app.vi_state.input_mode = InputMode.INSERT


@add_cmd(menu_title="Find")
def find() -> None:
    """Enter search mode."""
    start_global_search(direction=SearchDirection.FORWARD)


def find_prev_next(direction: SearchDirection) -> None:
    """Find the previous or next search match."""
    app = get_app()
    layout = app.layout
    control = app.layout.current_control
    # Determine search buffer and searched buffer
    search_buffer_control = None
    if isinstance(control, SearchBufferControl):
        search_buffer_control = control
        control = layout.search_links[search_buffer_control]
    elif isinstance(control, BufferControl):
        if control.search_buffer_control is not None:
            search_buffer_control = control.search_buffer_control
    elif app.search_bar is not None:
        search_buffer_control = app.search_bar.control
    if isinstance(control, BufferControl) and search_buffer_control is not None:
        # Update search_state.
        search_state = control.search_state
        search_state.direction = direction
        # Apply search to buffer
        buffer = control.buffer
        buffer.apply_search(search_state, include_current_position=False, count=1)
        # Set selection
        buffer.selection_state = SelectionState(
            buffer.cursor_position + len(search_state.text)
        )
        buffer.selection_state.enter_shift_mode()


@add_cmd()
def find_next() -> None:
    """Find the next search match."""
    find_prev_next(SearchDirection.FORWARD)


@add_cmd()
def find_previous() -> None:
    """Find the previous search match."""
    find_prev_next(SearchDirection.BACKWARD)


@add_cmd(
    filter=is_searching,
)
def stop_search() -> None:
    """Abort the search."""
    layout = get_app().layout
    buffer_control = layout.search_target_buffer_control
    if buffer_control is None:
        return
    search_buffer_control = buffer_control.search_buffer_control
    # Focus the previous control
    layout.focus(layout.previous_control)
    # Close the search toolbar
    if search_buffer_control is not None:
        del layout.search_links[search_buffer_control]
        # Reset content of search control.
        search_buffer_control.buffer.reset()
    # Redraw everything
    get_app().refresh()


@add_cmd(
    name="accept-search",
    filter=is_searching,
)
def accept_search() -> None:
    """Accept the search input."""
    layout = get_app().layout
    search_buffer_control = layout.current_control
    if not isinstance(search_buffer_control, BufferControl):
        log.debug("Current control not a buffer control")
        return
    for control in layout.find_all_controls():
        if (
            isinstance(control, BufferControl)
            and control.search_buffer_control == search_buffer_control
        ):
            search_state = control.search_state
            # Update search state.
            if search_buffer_control.buffer.text:
                search_state.text = search_buffer_control.buffer.text
            # Apply search.
            control.buffer.apply_search(
                search_state, include_current_position=True, count=1
            )

    # Set selection on target control
    buffer_control = layout.search_target_buffer_control
    if buffer_control and control.is_focusable():
        buffer = buffer_control.buffer
        buffer.selection_state = SelectionState(
            buffer.cursor_position + len(search_state.text)
        )
        buffer.selection_state.enter_shift_mode()

    # Add query to history of search line.
    search_buffer_control.buffer.append_to_history()
    # Stop the search
    stop_search()
